/*
 * "hello world" program:
 *
 * command-line argument gives number of threads.
 */
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

void * thread_fcn(void * thread_arg);

/* ---- main program ---- */

int main(int argc, char* argv[]) {

    char* usage_fmt = "usage:  %s number_of_threads\n";

    /* process command-line arguments */
    char* end_ptr_for_strtol;
    if (argc != 2)  {
        fprintf(stderr, usage_fmt, argv[0]);
        exit(EXIT_FAILURE);
    }
    int num_threads = strtol(argv[1], &end_ptr_for_strtol, 10);
    if (*end_ptr_for_strtol != '\0') {
        fprintf(stderr, usage_fmt, argv[0]);
        exit(EXIT_FAILURE);
    }

    /* 
     * set up IDs for threads (need a separate variable for since the
     * only way to pass data to a thread is via a pointer)
     */
    int threadIDs[num_threads];
    for (int i = 0; i < num_threads; ++i)
        threadIDs[i] = i;

    /* start num_threads new threads, each with different ID */
    pthread_t threads[num_threads];
    for (int i = 0; i < num_threads; ++i) 
        pthread_create(&threads[i], NULL, thread_fcn, (void *) &threadIDs[i]);

    printf("all threads launched\n");

    /* wait for all threads to complete */
    for (int i = 0; i < num_threads; ++i) 
        pthread_join(threads[i], NULL);

    printf("all threads complete\n");

    return EXIT_SUCCESS;
}

/* ---- code to be executed by each thread ---- */

void * thread_fcn(void * thread_arg) {
    int myID = * (int *) thread_arg;

    fprintf(stdout, "hello, world, from thread %d\n", myID);

    pthread_exit((void* ) NULL);
}
